@@ -14,7 +14,7 @@
  * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
  */
 
-/* $Id: pfflowd.c,v 1.18 2004/09/06 12:25:57 djm Exp $ */
+/* $Id: pfflowd.c,v 1.20 2006/06/07 10:58:37 msf Exp $ */
 
 #include <sys/types.h>
 #include <sys/ioctl.h>
@@ -229,7 +229,7 @@ format_pf_host(char *buf, size_t n, struct pf_state_host *h, sa_family_t af)
 }
 
 static int
-send_netflow_v1(const struct _PFSYNC_STATE *st, u_int n, int *flows_exp)
+send_netflow_v1(const struct pfsync_state *st, u_int n, int *flows_exp)
 {
 	char now_s[64];
 	int i, j, offset, num_packets, err;
@@ -308,13 +308,13 @@ send_netflow_v1(const struct _PFSYNC_STATE *st, u_int n, int *flows_exp)
 		}
 
 		flw = (struct NF1_FLOW *)(packet + offset);
-		if (netflow_socket != -1 && st[i].packets[0] != 0) {
+		if (netflow_socket != -1 && st[i].packets[0][0] != 0) {
 			flw->src_ip = src.addr.v4.s_addr;
 			flw->dest_ip = dst.addr.v4.s_addr;
 			flw->src_port = src.port;
 			flw->dest_port = dst.port;
-			flw->flow_packets = st[i].packets[0];
-			flw->flow_octets = st[i].bytes[0];
+			flw->flow_packets = st[i].packets[0][0];
+			flw->flow_octets = st[i].bytes[0][0];
 			flw->flow_start = htonl(uptime_ms - creation);
 			flw->flow_finish = htonl(uptime_ms);
 			flw->protocol = st[i].proto;
@@ -324,13 +324,13 @@ send_netflow_v1(const struct _PFSYNC_STATE *st, u_int n, int *flows_exp)
 			hdr->flows++;
 		}
 		flw = (struct NF1_FLOW *)(packet + offset);
-		if (netflow_socket != -1 && st[i].packets[1] != 0) {
+		if (netflow_socket != -1 && st[i].packets[1][0] != 0) {
 			flw->src_ip = dst.addr.v4.s_addr;
 			flw->dest_ip = src.addr.v4.s_addr;
 			flw->src_port = dst.port;
 			flw->dest_port = src.port;
-			flw->flow_packets = st[i].packets[1];
-			flw->flow_octets = st[i].bytes[1];
+			flw->flow_packets = st[i].packets[1][0];
+			flw->flow_octets = st[i].bytes[1][0];
 			flw->flow_start = htonl(uptime_ms - creation);
 			flw->flow_finish = htonl(uptime_ms);
 			flw->protocol = st[i].proto;
@@ -342,10 +342,10 @@ send_netflow_v1(const struct _PFSYNC_STATE *st, u_int n, int *flows_exp)
 		flw = (struct NF1_FLOW *)(packet + offset);
 
 		if (verbose_flag) {
-			packets_out = ntohl(st[i].packets[0]);
-			packets_in = ntohl(st[i].packets[1]);
-			bytes_out = ntohl(st[i].bytes[0]);
-			bytes_in = ntohl(st[i].bytes[1]);
+			packets_out = ntohl(st[i].packets[0][0]);
+			packets_in = ntohl(st[i].packets[1][0]);
+			bytes_out = ntohl(st[i].bytes[0][0]);
+			bytes_in = ntohl(st[i].bytes[1][0]);
 
 			creation_tt = now - (creation / 1000);
 			localtime_r(&creation_tt, &creation_tm);
@@ -399,8 +399,9 @@ send_netflow_v1(const struct _PFSYNC_STATE *st, u_int n, int *flows_exp)
 	return (ntohs(hdr->flows));
 }
 
+
 static int
-send_netflow_v5(const struct _PFSYNC_STATE *st, u_int n, int *flows_exp)
+send_netflow_v5(const struct pfsync_state *st, u_int n, int *flows_exp)
 {
 	char now_s[64];
 	int i, j, offset, num_packets, err;
@@ -480,13 +481,13 @@ send_netflow_v5(const struct _PFSYNC_STATE *st, u_int n, int *flows_exp)
 		}
 
 		flw = (struct NF5_FLOW *)(packet + offset);
-		if (netflow_socket != -1 && st[i].packets[0] != 0) {
+		if (netflow_socket != -1 && st[i].packets[0][0] != 0) {
 			flw->src_ip = src.addr.v4.s_addr;
 			flw->dest_ip = dst.addr.v4.s_addr;
 			flw->src_port = src.port;
 			flw->dest_port = dst.port;
-			flw->flow_packets = st[i].packets[0];
-			flw->flow_octets = st[i].bytes[0];
+			flw->flow_packets = st[i].packets[0][0];
+			flw->flow_octets = st[i].bytes[0][0];
 			flw->flow_start = htonl(uptime_ms - creation);
 			flw->flow_finish = htonl(uptime_ms);
 			flw->tcp_flags = 0;
@@ -496,13 +497,13 @@ send_netflow_v5(const struct _PFSYNC_STATE *st, u_int n, int *flows_exp)
 			hdr->flows++;
 		}
 		flw = (struct NF5_FLOW *)(packet + offset);
-		if (netflow_socket != -1 && st[i].packets[1] != 0) {
+		if (netflow_socket != -1 && st[i].packets[1][0] != 0) {
 			flw->src_ip = dst.addr.v4.s_addr;
 			flw->dest_ip = src.addr.v4.s_addr;
 			flw->src_port = dst.port;
 			flw->dest_port = src.port;
-			flw->flow_packets = st[i].packets[1];
-			flw->flow_octets = st[i].bytes[1];
+			flw->flow_packets = st[i].packets[1][0];
+			flw->flow_octets = st[i].bytes[1][0];
 			flw->flow_start = htonl(uptime_ms - creation);
 			flw->flow_finish = htonl(uptime_ms);
 			flw->tcp_flags = 0;
@@ -514,10 +515,10 @@ send_netflow_v5(const struct _PFSYNC_STATE *st, u_int n, int *flows_exp)
 		flw = (struct NF5_FLOW *)(packet + offset);
 
 		if (verbose_flag) {
-			packets_out = ntohl(st[i].packets[0]);
-			packets_in = ntohl(st[i].packets[1]);
-			bytes_out = ntohl(st[i].bytes[0]);
-			bytes_in = ntohl(st[i].bytes[1]);
+			packets_out = ntohl(st[i].packets[0][0]);
+			packets_in = ntohl(st[i].packets[1][0]);
+			bytes_out = ntohl(st[i].bytes[0][0]);
+			bytes_in = ntohl(st[i].bytes[1][0]);
 
 			creation_tt = now - (creation / 1000);
 			localtime_r(&creation_tt, &creation_tm);
@@ -571,6 +572,32 @@ send_netflow_v5(const struct _PFSYNC_STATE *st, u_int n, int *flows_exp)
 	return (ntohs(hdr->flows));
 }
 
+static void
+send_flow(const struct pfsync_state *st, u_int n, int *flows_exp)
+{
+	int r = 0;
+
+	switch (export_version) {
+	case 1:
+		r = send_netflow_v1(st, n, flows_exp);
+		break;
+	case 5:
+		r = send_netflow_v5(st, n, flows_exp);
+		break;
+	default:
+		/* should never reach this point */
+		syslog(LOG_DEBUG, "Invalid netflow version, exiting");
+		exit(1);
+	}
+
+	if (r > 0) {
+		flows_exported += r;
+		if (verbose_flag)
+			syslog(LOG_DEBUG, "flows_exported = %d", *flows_exp);
+	}
+
+}
+
 /*
  * Per-packet callback function from libpcap. 
  */
@@ -579,8 +606,8 @@ packet_cb(u_char *user_data, const struct pcap_pkthdr* phdr,
     const u_char *pkt)
 {
 	const struct pfsync_header *ph = (const struct pfsync_header *)pkt;
-	const struct _PFSYNC_STATE *st;
-	int r = 0;
+	const struct pfsync_state *st;
+	u_int64_t bytes[2], packets[2]; 
 
 	if (phdr->caplen < PFSYNC_HDRLEN) {
 		syslog(LOG_WARNING, "Runt pfsync packet header");
@@ -603,25 +630,56 @@ packet_cb(u_char *user_data, const struct pcap_pkthdr* phdr,
 		return;
 	}
 
-	st = (const struct _PFSYNC_STATE *)((const u_int8_t *)ph + sizeof(*ph));
+	st = (const struct pfsync_state *)((const u_int8_t *)ph + sizeof(*ph));
 
-	switch (export_version) {
-	case 1:
-		r = send_netflow_v1(st, ph->count, &flows_exported);
-		break;
-	case 5:
-		r = send_netflow_v5(st, ph->count, &flows_exported);
-		break;
-	default:
-		/* should never reach this point */
-		syslog(LOG_DEBUG, "Invalid netflow version, exiting");
-		exit(1);
-	}
+	/*
+	 * Check if any members of st->packets or st->bytes overflow
+	 * the 32 bit netflow counters, if so, create as many flow records 
+	 * that are needed to clear the counter. 
+	 */
+	 
+	 pf_state_counter_ntoh(st->packets[0],packets[0]); 
+	 pf_state_counter_ntoh(st->packets[1],packets[1]); 
+	 pf_state_counter_ntoh(st->bytes[0],bytes[0]); 
+	 pf_state_counter_ntoh(st->bytes[1],bytes[1]); 
 
-	if (r > 0) {
-		flows_exported += r;
-		if (verbose_flag)
-			syslog(LOG_DEBUG, "flows_exported = %d", flows_exported);
+	 while (bytes[0] > 0 || bytes[1] > 0 || 
+	        packets[0] > 0 || packets[1] > 0) {
+		
+		struct pfsync_state st1;
+
+		memcpy(&st1, st, sizeof(st1));
+
+		if (bytes[0] > UINT_MAX) {
+			st1.bytes[0][0] = 0xffffffff;
+			bytes[0] -= MIN(bytes[0], 0xffffffff);
+		} else {
+			st1.bytes[0][0] = htonl(bytes[0]);
+			bytes[0] = 0;
+		}
+		if (bytes[1] > UINT_MAX) {
+			st1.bytes[1][0] = 0xffffffff;
+			bytes[1] -= MIN(bytes[1], 0xffffffff);
+		} else {
+			st1.bytes[1][0] = htonl(bytes[1]);
+			bytes[1] = 0;
+		}
+		if (packets[0] > UINT_MAX) {
+			st1.packets[0][0] = 0xffffffff;
+			packets[0] -= MIN(packets[0], 0xffffffff);
+		} else {
+			st1.packets[0][0] = htonl(packets[0]);
+			packets[0] = 0;
+		}
+		if (packets[1] > UINT_MAX) {
+			st1.packets[1][0] = 0xffffffff;
+			packets[1] -= MIN(packets[1], 0xffffffff);
+		} else {
+			st1.packets[1][0] = htonl(packets[1]);
+			packets[1] = 0;
+		}
+
+		send_flow(&st1, ph->count, &flows_exported);
 	}
 }
 
@@ -807,8 +865,6 @@ main(int argc, char **argv)
 			exit(1);
 	}
 
-fprintf(stderr, "ZZZZ %d\n", netflow_socket);
-
 	if (dontfork_flag) {
 		if (!verbose_flag)
 			drop_privs();
